<?php
/*------------------------------------------------------------------------------
** File:        B4XSerializator.php
** Class:       B4XSerializator
** Description: Just a php port/implentation of B4XSerializator by Erel (Anywhere Software) from this page for the .Net Framework: https://www.b4x.com/android/forum/threads/b4x-net-framework-implementation-of-b4xserializator.73080/
** Version:     0.5
** Updated:     16.08.2019
** Author:      Toky Olivier RAZANAKOTONARIVO
** Homepage:    www.soinala-studio.com 
**------------------------------------------------------------------------------ */


include "lib/BinaryStream/BinaryStream.php";

class B4XSerializator{
	const T_NULL = 0;
	const T_STRING = 1;
	const T_SHORT = 2;
	const T_INT = 3;
	const T_LONG = 4;
	const T_FLOAT = 5;
	const T_DOUBLE = 6;
	const T_BOOLEAN = 7;
	const T_BYTE = 10;
	const T_CHAR = 14;
	const T_MAP = 20;
	const T_LIST = 21;
	const T_NSARRAY = 22;
	const T_NSDATA = 23;
	const T_TYPE = 24;
	
	protected $memorystream;
	protected $bs;
	
	public function __construct(){
		
	}
	
	public function ConvertBytesToObject($bytes) {
		$this->memorystream = fopen('php://memory', 'w+');
		$this->bs = new BinaryStream($this->memorystream, BinaryStream::CREATE);
		$this->bs->setEndian(BinaryStream::LITTLE);
		fwrite($this->memorystream, gzuncompress($bytes));
		rewind($this->memorystream);
		return $this->readObject();
	}
	
	public function ConvertObjectToBytes($object) {
		$this->memorystream = fopen('php://memory', 'w+');
		$this->bs = new BinaryStream($this->memorystream, BinaryStream::CREATE);
		$this->bs->setEndian(BinaryStream::LITTLE);
		
		$this->writeObject($object);
		rewind($this->memorystream);
		$bytes = gzencode(stream_get_contents($this->memorystream), -1, FORCE_DEFLATE);
		return $bytes;
	}
	
	public function B4XBytes($str) {
		return new B4XByteArray($str);
	}
	
	private function writeObject($o) {
		if (is_null($o))
			$this->writeByte(self::T_NULL);
		elseif (is_integer($o)) {
			$this->writeByte(self::T_INT);
			$this->writeInt($o);
		} elseif (is_float($o)) {
			$this->writeByte(self::T_FLOAT);
			$this->writeFloat($o);
		} elseif (is_double($o)) {
			$this->writeByte(self::T_DOUBLE);
			$this->writeDouble($o);
		} elseif (is_bool($o)) {
			$this->writeByte(self::T_BOOLEAN);
			$this->writeByte($o?1:0);
		} elseif (is_string($o)) {
			if (strlen($o)==1) {
				$this->writeByte(self::T_CHAR);
				$this->writeShort(ord($o[0]));
			} else {
				$this->writeByte(self::T_STRING);
				$this->writeInt(strlen($o));
				$this->writeStr($o);
			}
		} elseif (is_array($o)) {
			if ($this->isAssoc($o)) {
				$this->writeByte(self::T_MAP);
				$this->writeMap($o);
			} else {
				$this->writeByte(self::T_LIST);
				$this->writeList($o);
			}
		} elseif (is_a($o, "B4XType")) {
			$this->writeByte(self::T_TYPE);
			$this->writeType($o);
		} elseif (is_a($o, "B4XByteArray")) {
			$this->writeByte(self::T_NSDATA);
			$this->writeByteArray($o);
		} elseif (is_a($o, "B4XNumeric")) {
			$this->writeNumeric($o);
		}
	}
	
	private function writeByte($b) {
		$this->bs->writeInteger($b, 8);
	}
	
	private function writeShort($int) {
		$this->bs->writeInteger($int, 16);
	}
	
	private function writeInt($i) {
		$this->bs->writeInteger($i, 32);
	}
	
	private function writeLong($i) {
		$this->bs->writeInteger($i, 64);
	}
	
	private function writeStr($str) {
		$this->bs->writeString($str);
	}
	
	private function writeFloat($n) {
		$this->bs->writeFloat($n, 32);
	}
	
	private function writeDouble($n) {
		$this->bs->writeFloat($n, 64);
	}
	
	private function writeNumeric($obj) {
		$v = $obj->Value;
		switch ($obj->Type) {
			case self::T_BYTE:
				$this->writeByte(self::T_BYTE);
				$this->writeByte($v);
				break;
				
			case self::T_SHORT:
				$this->writeByte(self::T_SHORT);
				$this->writeShort($v);
				break;
			
			case self::T_INT:
				$this->writeByte(self::T_INT);
				$this->writeInt($v);
				break;
			
			case self::T_LONG:
				$this->writeByte(self::T_LONG);
				$this->writeLong($v);
				break;
				
			case self::T_FLOAT:
				$this->writeByte(self::T_FLOAT);
				$this->writeFloat($v);
				break;
		
			case self::T_DOUBLE:
				$this->writeByte(self::T_DOUBLE);
				$this->writeDouble($v);
				break;
		} 
	}
	
	private function writeList($arr) {
		$this->writeInt(sizeof($arr));
		foreach ($arr as $obj) {
			$this->writeObject($obj);
		}
	}
	
	private function writeMap($arr) {
		$this->writeInt(sizeof($arr));
		foreach ($arr as $key=>$value) {
			$this->writeObject($key);
			$this->writeObject($value);
		}
	}
	
	private function writeType($o) {
		$this->writeObject('_' . $o->ClassName);
		$this->writeMap($o->Fields);
	}
	
	private function writeByteArray($o) {
		$this->writeInt(strlen($o->Data));
		$this->writeStr($o->Data);
	}
	
	private function readObject() {
		$t = $this->readByte();
		switch ($t) {
			case self::T_NULL:
				return null;
			case self::T_BYTE:
				return $this->readByte();
			case self::T_INT:
				return $this->readInt();
			case self::T_LONG:
				return $this->readLong();
			case self::T_SHORT:
				return $this->readShort();
			case self::T_FLOAT:
				return $this->readFloat();
			case self::T_DOUBLE:
				return $this->readDouble();
			case self::T_BOOLEAN:
				return $this->readByte()==1?true:false;
			case self::T_CHAR:
				return chr($this->readShort());
			case self::T_STRING:
			case self::T_NSDATA:
				$len = $this->readInt();
				return $this->readStr($len);
			case self::T_LIST:
			case self::T_NSARRAY:
				return $this->readList();
			case self::T_MAP:
				return $this->readMap();
			case self::T_TYPE:
				return $this->readType();
		}
	}
	
	private function readByte() {
		return $this->bs->readInteger(8);
	}
	
	private function readShort(){
		return $this->bs->readInteger(16);
	}
	
	private function readInt(){
		return $this->bs->readInteger(32);
	}
	
	private function readLong(){
		return $this->bs->readInteger(64);
	}
	
	private function readStr($len){
		return $this->bs->readString($len);
	}
	
	private function readFloat(){
		return $this->bs->readFloat();
	}
	
	private function readDouble(){
		return $this->bs->readFloat(64);
	}
	
	private function readList(){
		$len = $this->readInt();
		$arr = array();
		for ($i=0;$i<$len;$i++) {
			$arr []=$this->readObject();
		}
		return $arr;
	}
	
	private function readMap(){
		$len = $this->readInt();
		$arr = array();
		for ($i=0;$i<$len;$i++) {
			$key = $this->readObject();
			$value = $this->readObject();
			$arr [$key]=$value;
		}
		return $arr;
	}
	
	private function readType(){
		$cls = $this->readObject();
		$data = $this->readMap();
		return new B4XType($cls, $data);
	}
	
	private function is_byte($int) {
		return ($int>=0 && $int<=255)?true:false;
	}
	
	private function is_short($int) {
		return ($int>=-32768 && $int<=32767)?true:false;
	}
	
	private function isAssoc(array $arr)
	{
		if (array() === $arr) return false;
		return array_keys($arr) !== range(0, count($arr) - 1);
	}
	
	public function OutputBytes($data) {
		$_size = strlen($data);
		header('Content-Type: binary/octet-stream');
		header('Content-Length: '.$_size);
		echo $data;
	}

}

class B4XType {
	public $ClassName;
	public $Fields;
	public function __construct($className, $data){
		$i = strrpos($className, '$');
		if ($i>-1) {
			$className = substr($className, $i+2);
		} elseif ($className[0]=='_') {
			$className = substr($className, 1);
		} else {
			$className = strtolower($className);
		}
		
		$this->ClassName = $className;
		$this->Fields = $data;
	}
}

class B4XByteArray {
	public $Data;
	public function __construct($data){
		$this->Data = $data;
	}
}

class B4XNumeric {
	public $Value;
	public $Type;
	public function __construct($n, $type){
		$this->Value = $n;
		$this->Type = $type;
	}
}
